package ru.dbotthepony.mc.otm.registry

import com.mojang.serialization.Codec
import net.minecraft.core.UUIDUtil
import net.minecraft.core.component.DataComponentType
import net.minecraft.core.registries.BuiltInRegistries
import net.minecraft.network.RegistryFriendlyByteBuf
import net.minecraft.network.codec.StreamCodec
import net.neoforged.bus.api.IEventBus
import net.neoforged.neoforge.fluids.FluidStack
import ru.dbotthepony.mc.otm.capability.matter.PatternState
import ru.dbotthepony.mc.otm.container.ItemFilter
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.data.DecimalCodec
import ru.dbotthepony.mc.otm.network.StreamCodecs
import java.util.UUID

object MDataComponentTypes {
	private val registry = MDeferredRegister(BuiltInRegistries.DATA_COMPONENT_TYPE)

	// class instead of object to preserve identity distinction for registry reversal search
	private class DecimalComponent : DataComponentType<Decimal> {
		override fun codec(): Codec<Decimal> {
			return DecimalCodec
		}

		override fun streamCodec(): StreamCodec<in RegistryFriendlyByteBuf, Decimal> {
			return DecimalCodec.NETWORK
		}
	}

	val FLUID_STACK by registry.register("fluid_stack") {
		object : DataComponentType<FluidStack> {
			override fun codec(): Codec<FluidStack> {
				return FluidStack.OPTIONAL_CODEC
			}

			override fun streamCodec(): StreamCodec<in RegistryFriendlyByteBuf, FluidStack> {
				return FluidStack.OPTIONAL_STREAM_CODEC
			}
		}
	}

	private fun uuid() = DataComponentType.builder<UUID>().persistent(UUIDUtil.CODEC).networkSynchronized(UUIDUtil.STREAM_CODEC).build()

	val BATTERY_LEVEL: DataComponentType<Decimal> by registry.register("battery_level") { DecimalComponent() }
	val MAX_BATTERY_LEVEL: DataComponentType<Decimal> by registry.register("max_battery_level") { DecimalComponent() }
	val MAX_BATTERY_INPUT: DataComponentType<Decimal> by registry.register("max_battery_input") { DecimalComponent() }
	val MAX_BATTERY_OUTPUT: DataComponentType<Decimal> by registry.register("max_battery_output") { DecimalComponent() }
	val MATTER_LEVEL: DataComponentType<Decimal> by registry.register("matter_level") { DecimalComponent() }

	val EXOPACK_SLOT_COUNT: DataComponentType<Int> by registry.register("exopack_slot_count") { DataComponentType.builder<Int>().persistent(Codec.INT).networkSynchronized(StreamCodecs.INT).build() }
	val EXOPACK_UPGRADE_ID: DataComponentType<UUID> by registry.register("exopack_upgrade_id") { uuid() }
	val QUANTUM_LINK_ID: DataComponentType<UUID> by registry.register("quantum_link_id") { uuid() }
	val CONDENSATION_DRIVE_UUID: DataComponentType<UUID> by registry.register("condensation_drive_uuid") { uuid() }
	val PATTERNS: DataComponentType<List<PatternState>> by registry.register("patterns") { DataComponentType.builder<List<PatternState>>().persistent(Codec.list(PatternState.CODEC)).build() }
	val ITEM_FILTER: DataComponentType<ItemFilter> by registry.register("item_filter") { DataComponentType.builder<ItemFilter>().persistent(ItemFilter.CODEC).build() }

	val PRIMED by registry.register("primed") {
		object : DataComponentType<Boolean> {
			override fun codec(): Codec<Boolean> {
				return Codec.BOOL
			}

			override fun streamCodec(): StreamCodec<in RegistryFriendlyByteBuf, Boolean> {
				return StreamCodecs.BOOLEAN
			}
		}
	}

	fun register(bus: IEventBus) {
		registry.register(bus)
	}
}
